#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include "charset.h"
#include "string16.h"
#include "dtd.h"
#include "input.h"
#include "xmlparser.h"
#include "stdio16.h"

void print_bit(Parser p, XBit bit);
void print_attrs(ElementDefinition e, Attribute a);
void print_text(Char *text);
void print_text_bit(Char *text);
void dtd_cb(XBit bit, void *arg);

int verbose = 0, expand = 0, bits = 0, silent = 0, nsgml = 0,
    attr_defaults = 0, merge = 0, strict_xml = 0;
char *enc_name = 0;
CharacterEncoding encoding = CE_unknown, requested_encoding = CE_unknown;
InputSource source = 0;

int main(int argc, char **argv)
{
    int i;
    Parser p;
    char *s;

    /* Sigh... guess which well-known system doesn't have getopt() */

    for(i = 1; i < argc; i++)
    {
	if(argv[i][0] != '-')
	    break;
	for(s = &argv[i][1]; *s; s++)
	    switch(*s)
	    {
	    case 'v':
		verbose = 1;
		break;
	    case 'a':
		attr_defaults = 1;
		break;
	    case 'e':
		expand = 1;
		break;
	    case 'b':
		bits = 1;
		break;
	    case 's':
		silent = 1;
		break;
	    case 'n':
		nsgml = 1;
		break;
	    case 'c':
		enc_name = argv[++i];
		break;
	    case 'm':
		merge = 1;
		break;
	    case 'x':
		strict_xml = 1;
		attr_defaults = 1;
		expand = 1;
		break;
	    default:
		fprintf(stderr, 
			"usage: rxp [-abemnsvx] [-c encoding] [url]\n");
		return 1;
	    }
    }

    if(i < argc)
    {
	Entity ent = NewExternalEntity(0, 0, argv[i], 0, 0);
	if(ent)
	    source = EntityOpen(ent);
    }
    else
	source = SourceFromStream("<stdin>", stdin);

    if(!source)
	return 1;

    p = NewParser();

    if(bits)
    {
	ParserSetDtdCallback(p, dtd_cb);
	ParserSetCallbackArg(p, p);
    }

    if(attr_defaults)
	ParserSetFlag(p, ReturnDefaultedAttributes, 1);

    if(!expand)
    {
	ParserSetFlag(p, ExpandGeneralEntities, 0);
	ParserSetFlag(p, ExpandCharacterEntities, 0);
    }

    if(merge)
	ParserSetFlag(p, MergePCData, 1);

    if(nsgml)
    {
	ParserSetFlag(p, XMLPiEnd, 0);
	ParserSetFlag(p, XMLEmptyTagEnd, 0);
	ParserSetFlag(p, XMLPredefinedEntities, 0);
	ParserSetFlag(p, XMLExternalIDs, 0);
	ParserSetFlag(p, XMLMiscWFWarnings, 0);
	ParserSetFlag(p, TrustSDD, 0);
	ParserSetFlag(p, WarnOnUnquotedAttributeValues, 0);
	ParserSetFlag(p, ExpandGeneralEntities, 0);
	ParserSetFlag(p, ExpandCharacterEntities, 0);
/*	ParserSetFlag(p, TrimPCData, 1); */
    }

    if(strict_xml)
    {
	ParserSetFlag(p, ErrorOnBadCharacterEntities, 1);	
	ParserSetFlag(p, ErrorOnUndefinedEntities, 1);
	ParserSetFlag(p, ErrorOnXMLWFErrors, 1);
	ParserSetFlag(p, WarnOnUndefinedElements, 0);
	ParserSetFlag(p, WarnOnUndefinedAttributes, 0);
	ParserSetFlag(p, WarnOnRedefinitions, 0);
	
    }

    if(ParserPush(p, source) == -1)
    {
	ParserPerror(p, &p->xbit);
	return 1;
    }

    if(enc_name)
    {
	for(i=0; i<CE_enum_count; i++)
	    if(strcasecmp8(enc_name, CharacterEncodingNameAndByteOrder[i]) == 0)
	    {
		requested_encoding = i;
		break;
	    }
	if(requested_encoding == CE_unknown)
	    for(i=0; i<CE_enum_count; i++)
		if(strcasecmp8(enc_name, CharacterEncodingName[i]) == 0)
		{
		    requested_encoding = i;
		    break;
		}
	if(requested_encoding == CE_unknown)
	{
	    fprintf(stderr, "unknown encoding %s\n", enc_name);
	    return 1;
	}
	encoding = requested_encoding;
    }
    else if(strict_xml)
	encoding = CE_UTF_8;
    else
	encoding = source->entity->encoding;

    if(verbose)
	fprintf(stderr, "Input encoding %s, output encoding %s\n",
		CharacterEncodingNameAndByteOrder[source->entity->encoding],
		CharacterEncodingNameAndByteOrder[encoding]);

    if(!strict_xml && source->entity->ml_decl == ML_xml && !bits)
    {
	Printf(encoding, "<?xml");
	if(source->entity->version_decl)
	    Printf(encoding, " version=\"%s\"", 
		   source->entity->version_decl);
	if(source->entity->encoding_decl)
	    Printf(encoding, " encoding=\"%s\"", 
		   source->entity->encoding_decl);
	if(source->entity->standalone_decl)
	    Printf(encoding, " standalone=\"%s\"", 
		   source->entity->standalone_decl);
	Printf(encoding, "?>\n");
    }

    while(1)
    {
	XBit bit = ReadXBit(p);
	print_bit(p, bit);
	if(bit->type == XBIT_eof)
	    return 0;
	if(bit->type == XBIT_error)
	    return 1;
	/* XXX handle encoding in XML bit */
	FreeXBit(bit);
    }
}

void print_bit(Parser p, XBit bit)
{
    const char *sys, *pub;

    if(silent && bit->type != XBIT_error)
	return;

    if(bits)
    {
	Printf(encoding, "At %d: ", bit->byte_offset);
	switch(bit->type)
	{
	case XBIT_eof:
	    Printf(encoding, "EOF\n");
	    break;
	case XBIT_error:
	    ParserPerror(p, bit);
	    break;
	case XBIT_dtd:
	    sys = pub = "<none>";
	    if(bit->dtd_external_part)
	    {
		if(bit->dtd_external_part->publicid)
		    pub = bit->dtd_external_part->publicid;
		if(bit->dtd_external_part->systemid)
		    sys = bit->dtd_external_part->systemid;
	    }
	    Printf(encoding, "doctype: %S pubid %s sysid %s\n", 
		   bit->dtd_name, pub, sys);
	    break;
	case XBIT_start:
	    Printf(encoding, "start: %S ", bit->element_definition->name);
	    print_attrs(0, bit->attributes);
	    Printf(encoding, "\n");
	    break;
	case XBIT_empty:
	    Printf(encoding, "empty: %S ", bit->element_definition->name);
	    print_attrs(0, bit->attributes);
	    Printf(encoding, "\n");
	    break;
	case XBIT_end:
	    Printf(encoding, "end: %S\n", bit->element_definition->name);
	    break;
	case XBIT_pi:
	    Printf(encoding, "pi: %S: ", bit->pi_name);
	    print_text_bit(bit->pi_chars);
	    Printf(encoding, "\n");
	    break;
	case XBIT_cdsect:
	    Printf(encoding, "cdata: ");
	    print_text_bit(bit->cdsect_chars);
	    Printf(encoding, "\n");
	    break;
	case XBIT_pcdata:
	    Printf(encoding, "pcdata: ");
	    print_text_bit(bit->pcdata_chars);
	    Printf(encoding, "\n");
	    break;
	case XBIT_comment:
	    Printf(encoding, "comment: ");
	    print_text_bit(bit->comment_chars);
	    Printf(encoding, "\n");
	    break;
	default:
	    fprintf(stderr, "***%s\n", XBitTypeName[bit->type]);
	    exit(1);
	    break;
	}
    }
    else
    {
	switch(bit->type)
	{
	case XBIT_eof:
	    break;
	case XBIT_error:
	    ParserPerror(p, bit);
	    break;
	case XBIT_dtd:
	    if(strict_xml)
		/* no doctype in canonical XML */
		break;
	    Printf(encoding, "<!DOCTYPE %S", bit->dtd_name);
	    if(bit->dtd_external_part)
	    {
		if(bit->dtd_external_part->publicid)
		    Printf(encoding, " PUBLIC \"%s\"", 
			   bit->dtd_external_part->publicid);
		else if(bit->dtd_external_part->systemid)
		    Printf(encoding, " SYSTEM");
		if(bit->dtd_external_part->systemid)
		    Printf(encoding, " \"%s\"", 
			   bit->dtd_external_part->systemid);
	    }
	    if(bit->dtd_internal_part)
		Printf(encoding, " [%S]", bit->dtd_internal_part->text);
	    Printf(encoding, ">\n");
	    break;
	case XBIT_start:
	case XBIT_empty:
	    Printf(encoding, "<%S", bit->element_definition->name);
	    print_attrs(bit->element_definition, bit->attributes);
	    if(bit->type == XBIT_start)
		Printf(encoding, ">");
	    else if(strict_xml)
		Printf(encoding, "></%S>", bit->element_definition->name);
	    else
		Printf(encoding, "/>");
	    break;
	case XBIT_end:
	    Printf(encoding, "</%S>", bit->element_definition->name);
	    break;
	case XBIT_pi:
	    Printf(encoding, "<?%S %S%s", 
		   bit->pi_name, bit->pi_chars, nsgml ? ">" : "?>");
	    if(p->state == PS_prolog && !strict_xml)
		Printf(encoding, "\n");
	    break;
	case XBIT_cdsect:
	    if(strict_xml)
		/* Print CDATA sections as plain PCDATA in canonical XML */
		print_text(bit->cdsect_chars);
	    else
		Printf(encoding, "<![CDATA[%S]]>", bit->cdsect_chars);
	    break;
	case XBIT_pcdata:
	    print_text(bit->pcdata_chars);
	    break;
	case XBIT_comment:
	    if(strict_xml)
		/* no comments in canonical XML */
		break;
	    Printf(encoding, "<!--%S-->", bit->comment_chars);
	    if(p->state == PS_prolog)
		Printf(encoding, "\n");
	    break;
	default:
	    fprintf(stderr, "\n***%s\n", XBitTypeName[bit->type]);
	    exit(1);
	    break;
	}
    }
}

int attr_compare(const void *a, const void *b)
{
    return Strcmp((*(Attribute *)a)->definition->name,
		  (*(Attribute *)b)->definition->name);
}

void print_attrs(ElementDefinition e, Attribute a)
{
    Attribute b;
    Attribute *aa;
    int i, n = 0;
    
    for(b=a; b; b=b->next)
	n++;

    if(n == 0)
	return;

    aa = malloc(n * sizeof(*aa));

    for(i=0, b=a; b; i++, b=b->next)
	aa[i] = b;

    if(strict_xml)
	qsort((void *)aa, n, sizeof(*aa), attr_compare);

    for(i=0; i<n; i++)
    {
	Printf(encoding, " %S=\"", aa[i]->definition->name);
	print_text(aa[i]->value);
	Printf(encoding, "\"");
    }
}

void print_text_bit(Char *text)
{
    int i;

    for(i=0; i<50 && text[i]; i++)
	if(text[i] == '\n' || text[i] == '\r')
	    text[i] = '~';
    Printf(encoding, "%.50S", text);
}

void dtd_cb(XBit bit, void *arg)
{
    Printf(encoding, "In DTD: ");
    print_bit(arg, bit);
}
	
void print_text(Char *text)
{
    Char *pc, *last;
    
    if(bits)
    {
	Printf(encoding, "%S", text);
	return;
    }

    for(pc = last = text; *pc; pc++)
    {
	if(*pc == '&' || *pc == '<' || *pc == '>' || *pc == '"' ||
	   (strict_xml && (*pc == 9 || *pc == 10 || *pc == 13)))
	{
	    if(pc > last)
		Printf(encoding, "%.*S", pc - last, last);
	    switch(*pc)
	    {
	    case '<':
		Printf(encoding, "&lt;");
		break;
	    case '>':
		Printf(encoding, "&gt;");
		break;
	    case '&':
		Printf(encoding, "&amp;");
		break;
	    case '"':
		Printf(encoding, "&quot;");
		break;
	    case 9:
		Printf(encoding, "&#9;");
		break;
	    case 10:
		Printf(encoding, "&#10;");
		break;
	    case 13:
		Printf(encoding, "&#13;");
		break;
	    }
	    last = pc+1;
	}
    }
	
    if(pc > last)
	Printf(encoding, "%.*S", pc - last, last);
}
